import copy

import logging

from deepdiff import DeepDiff

def appendSystem(systemKey,sim,sim2app,mode):

    logger = logging.getLogger("pyUAMMD")

    #########################
    #Appending system
    #System structure:
    #   systemKey: {
    #       "information": {
    #           "type": ["Simulation","Information"],
    #           "parameters": {"name": simName,
    #                          ...}
    #       }
    #       "backup": {
    #           "type": ["Simulation","Backup"],
    #           "parameters": {...},
    #       }

    # Four cases:
    # 1) system NOT in sim and system NOT in sim2app
    # 2) system NOT in sim and system in sim2app
    # 3) system in sim and system NOT in sim2app
    # 4) system in sim and system in sim2app

    systemInSim     = (systemKey in sim)
    systemInSim2app = (systemKey in sim2app)

    # 1) system NOT in sim and system NOT in sim2app
    if not systemInSim and not systemInSim2app:
        #Do nothing
        pass
    # 2) system NOT in sim and system in sim2app
    elif not systemInSim and systemInSim2app:
        sim[systemKey] = copy.deepcopy(sim2app[systemKey])
    # 3) system in sim and system NOT in sim2app
    elif systemInSim and not systemInSim2app:
        #Do nothing
        pass
    # 4) system in sim and system in sim2app
    elif systemInSim and systemInSim2app:
        infoInSim   = False
        backupInSim = False
        for key in sim[systemKey]:
            tpy,subTpy = sim[systemKey][key]["type"]

            if tpy == "Simulation" and subTpy == "Information":
                if not infoInSim:
                    infoInSim    = True
                    infoInSimKey = key
                else:
                    logger.error("More than one information system in simulation!")
                    raise Exception("More than one information system in simulation!")
            elif tpy == "Simulation" and subTpy == "Backup":
                if not backupInSim:
                    backupInSim    = True
                    backupInSimKey = key
                else:
                    logger.error("More than one backup system in simulation!")
                    raise Exception("More than one backup system in simulation!")
            else:
                logger.error("Unknown system type in simulation!")
                raise Exception("Unknown system type!")

        infoInSim2app   = False
        backupInSim2app = False
        for key in sim2app[systemKey]:
            tpy,subTpy = sim2app[systemKey][key]["type"]

            if tpy == "Simulation" and subTpy == "Information":
                if not infoInSim2app:
                    infoInSim2app    = True
                    infoInSim2appKey = key
                else:
                    logger.error("More than one information system in simulation!")
                    raise Exception("More than one information system in simulation!")
            elif tpy == "Simulation" and subTpy == "Backup":
                if not backupInSim2app:
                    backupInSim2app    = True
                    backupInSim2appKey = key
                else:
                    logger.error("More than one backup system in simulation!")
                    raise Exception("More than one backup system in simulation!")
            else:
                logger.error("Unknown system type in simulation to append!")
                raise Exception("Unknown system type!")

        # Four cases for info:
        # 1) info NOT in sim and info NOT in sim2app
        # 2) info NOT in sim and info in sim2app
        # 3) info in sim and info NOT in sim2app
        # 4) info in sim and info in sim2app

        # 1) info NOT in sim and info NOT in sim2app
        if not infoInSim and not infoInSim2app:
            #Do nothing
            pass
        # 2) info NOT in sim and info in sim2app
        elif not infoInSim and infoInSim2app:
            sim[systemKey][infoInSim2appKey] = copy.deepcopy(sim2app[systemKey][infoInSim2appKey])
        # 3) info in sim and info NOT in sim2app
        elif infoInSim and not infoInSim2app:
            #Do nothing
            pass
        # 4) info in sim and info in sim2app
        elif infoInSim and infoInSim2app:

            nameInSim   = sim[systemKey][infoInSimKey]["parameters"].get("name",None)
            nameInSim2app = sim2app[systemKey][infoInSim2appKey]["parameters"].get("name",None)

            if nameInSim == None and nameInSim2app == None:
                #Do nothing
                pass
            elif nameInSim != None and nameInSim2app == None:
                #Do nothing
                pass
            elif nameInSim == None and nameInSim2app != None:
                sim[systemKey][infoInSimKey]["parameters"]["name"] = nameInSim2app
            elif nameInSim != None and nameInSim2app != None:
                if mode == "modelId":
                    #Name must be the same
                    self.logger.error("System name must be the same. (Mode: modelId)")
                    raise Exception("System name must be the same.")
                elif mode == "batchId":
                    #Combine names
                    sim[systemKey][infoInSimKey]["parameters"]["name"] += "_" + nameInSim2app

            ###################################################################################

            seedInSim     = sim[systemKey][infoInSimKey]["parameters"].get("seed",None)
            seedInSim2app = sim2app[systemKey][infoInSim2appKey]["parameters"].get("seed",None)

            if seedInSim == None and seedInSim2app == None:
                #Do nothing
                pass
            elif seedInSim != None and seedInSim2app == None:
                #Do nothing
                pass
            elif seedInSim == None and seedInSim2app != None:
                sim[systemKey][infoInSimKey]["parameters"]["seed"] = seedInSim2app
            elif seedInSim != None and seedInSim2app != None:
                if seedInSim != seedInSim2app:
                    logger.error("Seeds of two systems are different!")
                    raise Exception("Seeds of two systems are different!")

                sim[systemKey][infoInSimKey]["parameters"]["seed"] = seedInSim2app

        # Four cases for backup:
        # 1) backup NOT in sim and backup NOT in sim2app
        # 2) backup NOT in sim and backup in sim2app
        # 3) backup in sim and backup NOT in sim2app
        # 4) backup in sim and backup in sim2app

        # 1) backup NOT in sim and backup NOT in sim2app
        if not backupInSim and not backupInSim2app:
            #Do nothing
            pass
        # 2) backup NOT in sim and backup in sim2app
        elif not backupInSim and backupInSim2app:
            sim[systemKey][backupInSim2appKey] = copy.deepcopy(sim2app[systemKey][backupInSim2appKey])
        # 3) backup in sim and backup NOT in sim2app
        elif backupInSim and not backupInSim2app:
            #Do nothing
            pass
        # 4) backup in sim and backup in sim2app
        elif backupInSim and backupInSim2app:

            equal = not DeepDiff(sim[systemKey][backupInSimKey],
                                 sim2app[systemKey][backupInSim2appKey],
                                 ignore_order=True,report_repetition=True)

            if not equal:
                logger.error("Backup systems are different!")
                raise Exception("Backup systems are different!")

    #System appended
    #########################




import logging

import copy
import itertools

from tqdm import tqdm

def mergeSimulationsSet(simulationsSet,
                        independentSimulationSteps=True,
                        additionalIgnoredSimulationStepTypes=[]):

    ignoredSimulationStepTypes = ['UtilsStep','FlowControl','Groups'] + additionalIgnoredSimulationStepTypes

    logger = logging.getLogger("pyUAMMD")

    if independentSimulationSteps:
        #We assume that batchId is 0 (not set) for each independent simulation
        #Check it
        for sim in simulationsSet:
            structureLabels = sim["topology"]["structure"]["labels"]
            if "batchId" in structureLabels:
                logger.error("[pyUAMMD] BatchId already set, for a single simulation. Cannot aggregate simulations")
                raise Exception("BatchId already set")

        #Update simulationsStep for each simulation in the simulation set
        #This ensures the simulation step is applied over particles for the
        #same batchId
        for simIndex,sim in enumerate(simulationsSet):

            if "simulationStep" in sim.keys():
                groupDefinitionRequired = False

                keysToRename = []
                for simStep in sim["simulationStep"].keys():

                    simStepType = sim["simulationStep"][simStep]["type"][0]

                    isIgnored = False
                    for ignoredType in ignoredSimulationStepTypes:
                        if ignoredType in simStepType:
                            isIgnored = True
                            break

                    if not isIgnored:

                        #Each simulationStep apply to a batchId group
                        if "group" not in sim["simulationStep"][simStep]["parameters"].keys():
                            sim["simulationStep"][simStep]["parameters"]["group"] = f"batchId_{simIndex}"
                            groupDefinitionRequired = True

                        keysToRename.append(simStep)

                    if "Groups" in simStepType:

                        #We rename all groups declared
                        for i in range(len(sim["simulationStep"][simStep]["data"])):
                            oldName = sim["simulationStep"][simStep]["data"][i][0]
                            newName = oldName + f"_{simIndex}"
                            sim["simulationStep"][simStep]["data"][i][0] = newName

                            #Rename all references to the group
                            for simStep2rename in sim["simulationStep"].keys():
                                if "group" in sim["simulationStep"][simStep2rename]["parameters"].keys():
                                    if sim["simulationStep"][simStep2rename]["parameters"]["group"] == oldName:
                                        sim["simulationStep"][simStep2rename]["parameters"]["group"] = newName

                for k in keysToRename:
                    sim["simulationStep"][k+f"_{simIndex}"] = sim["simulationStep"].pop(k)

                if groupDefinitionRequired:
                    if "groups_batchId" not in sim["simulationStep"].keys():
                        sim["simulationStep"]["groups_batchId"] = {
                            "type": ["Groups","GroupsList"],
                            "parameters": {},
                            "labels":["name","type","selection"],
                            "data":[
                                [f"batchId_{simIndex}","BatchIds",[0]],
                            ]
                        }
                    else:
                        logger.error("[pyUAMMD] groups_batchId already defined")
                        raise Exception("groups_batchId already defined")

    aggregatedSimulation = None
    for sim in tqdm(simulationsSet,desc="Merging simulations set"):
        if aggregatedSimulation is None:
            aggregatedSimulation = copy.deepcopy(sim)
        else:
            aggregatedSimulation.append(sim,mode="batchId")

    #Check if more than one neighbor list is defined
    #List all neighbor lists
    neighborLists = []
    if "forceField" in aggregatedSimulation["topology"].keys():
        for entry in aggregatedSimulation["topology"]["forceField"].keys():
            if aggregatedSimulation["topology"]["forceField"][entry]["type"][0] == "VerletConditionalListSet":
                neighborLists.append(entry)

    if len(neighborLists) > 1:
        nlsTypes = ""
        for i in neighborLists:
            nlsTypes += f"\"{i}\" ({aggregatedSimulation['topology']['forceField'][i]['type'][1]}), "
        nlsTypes = nlsTypes[:-2]+"."
        logger.warning(f"[pyUAMMD] Resulsting simulation will have multiple neighbor lists !!!. Neighbor lists types: {nlsTypes}")

    return aggregatedSimulation

